unit mt4_timezone_support_seeklist_class;
// This class is used by the TZSupport-Class to cache
// timestamps calculated by StrToBrokerTime().
// It turned out that the historic time zone calculation
// involves a pretty expensive database call. In this way
// the database call is avoided when the (date, time, timezone)
// has already been calculated.
// The cache manages itself by a simple LRU algorithm.
{$mode objfpc}{$H+}

interface
uses
	CMem,
	Classes,
	Windows,
	DateUtils,
	SysUtils
;
type
	// Datastructure to cache a date/time/timezone result
        TDateSeekEntry = Class
                protected
	        	seekDate	: TDateTime;
	                seekTimeStr	: AnsiString;
                        seekTimeZone	: AnsiString;
	                lastAccessed	: TDateTime;
	                Timestamp	: LongInt;
                        procedure setAccessed;
                public
                        constructor create(
                        	pSeekDate	: TDateTime;
                                pSeekTimeStr	: AnsiString;
                                pBrokerTZ	: AnsiString;
                                pTimeStamp	: LongInt
                        );
                        destructor Destroy; override;
                        // gets the last accessed timestamp for the LRU.
                        function getLastAccessed 	: tDateTime;
                        // This functions get the seek parameters.
                        function getSeekDate		: TDateTime;
                        function getSeekTimeStr		: AnsiString;
                        function getSeekTimeZone	: AnsiString;
                        // This function gets the return timestamp.
                        function getTimeStamp		: LongInt;
        end;

        // Selforganizing list for to cache the TDateSeekEntry objects
        TDateSeekList=class(TList)
                protected
                        seekCacheAddCount	: Integer;
                        procedure Msg(ALevel: Integer; AMethod: Ansistring; AMessage: Ansistring);
			procedure Msg(ALevel: Integer; AMethod: Ansistring; AMessage: Ansistring; AArgs: array of const);
                public
                        constructor create;
                	destructor destroy; override;
                        // This function returns a PosixTimeStamp when
                        // successful and -1 if there had not been any
                        // match for the parameters.
	                function seekTimeStamp(
	                        	pSeekDate	: TDateTime;
	                                pSeekTimeStr	: AnsiString;
	                                pBrokerTZ	: AnsiString
	                ) : LongInt;
                        // This procedure adds a new date/time/timezone
                        // to the list and checks for LRU cleanup.
	                procedure addTimeStamp(
	                        	pSeekDate	: TDateTime;
	                                pSeekTimeStr	: AnsiString;
	                                pBrokerTZ	: AnsiString;
	                                pTimeStamp	: LongInt
	                );
        end;

implementation

constructor TDateSeekList.create();
begin
        inherited create();
        seekCacheAddCount:=0;
end;

destructor TDateSeekList.destroy();
var
	i		: Integer;
        t		: TDateSeekEntry;
begin
	i:=0;
        while (i<Count) and (Count>0) do begin
        	t := TDateSeekEntry(Items[i]);
                t.Destroy;
                delete(i);
        end;
        inherited destroy();
end;


procedure TDateSeekList.Msg(ALevel: Integer; AMethod: Ansistring; AMessage: Ansistring);
begin
	AMessage := '<' + IntToStr(ALevel) + '> ' + AMethod + ': ' + AMessage;
	{$ifdef win32}
	OutputDebugString(PChar(AMessage));
	{$else}
	writeln(AMessage);
	{$endif}
end;

procedure TDateSeekList.Msg(ALevel: Integer; AMethod: Ansistring; AMessage: Ansistring; AArgs: array of const);
begin
	try
		Msg(ALevel, AMethod, Format(AMessage, AArgs));
	except
		Msg(ALevel, AMethod, 'Could not format message' );
	end;
end;

function TDateSeekList.seekTimeStamp(
	pSeekDate	: TDateTime;
        pSeekTimeStr	: AnsiString;
        pBrokerTZ	: AnsiString
) : LongInt;
var
	i		: Integer;
        rc		: LongInt;
        t		: TDateSeekEntry;
begin
  	rc:=-1;
        i:=0;
        while ( (i<Count) and (rc=-1) ) do begin
		t:= TDateSeekEntry(Items[i]);
        	if (
                	(t.getSeekDate=pSeekDate) AND
                        (t.getSeekTimeStr=pSeekTimeStr) AND
                        (t.getSeekTimeZone=pBrokerTZ)
                ) then begin
                        rc:=t.getTimeStamp;
                end else inc(i);
        end;
  	exit(rc);
end;

procedure TDateSeekList.addTimeStamp(
	pSeekDate	: TDateTime;
        pSeekTimeStr	: AnsiString;
        pBrokerTZ	: AnsiString;
        pTimeStamp	: LongInt
);
var
	t		: TDateSeekEntry;
        i		: Integer;
begin
  	t := TDateSeekEntry.create(pSeekDate, pSeekTimeStr, pBrokerTZ, pTimeStamp);
        Add(t);
        Inc(SeekCacheAddCount);
        if (SeekCacheAddCount>50) then begin
                Msg(0, 'addTimeStamp',
	        	'Cleaning up cache count=%d',
                        [Count]
                );
                // Check whether there are timestamps which can be cleaned up
                i:=0;
                while ( (i<Count) and (Count>0) ) do begin
                	t:=TDateSeekEntry(Items[i]);
                        // Remove Entries which were not accessed the last 2 minutes.
                        if ( MinuteSpan(Now, t.getLastAccessed)>2 ) then begin
                                Delete(i);
                                t.Destroy;
                                // i:=i-1;
                        end else inc(i);
                end;
                seekCacheAddCount:=0;
                Msg(0, 'addTimeStamp',
	        	'Cleaned up cache count=%d',
                        [Count]
                );
        end;
end;


procedure TDateSeekEntry.setAccessed;
begin
	lastAccessed := Now;
end;

constructor TDateSeekEntry.create(
	pSeekDate	: TDateTime;
        pSeekTimeStr	: AnsiString;
        pBrokerTZ	: AnsiString;
        pTimeStamp	: LongInt
);
begin
        setAccessed();
        seekDate := pSeekDate;
        seekTimeStr := pSeekTimeStr;
        seekTimeZone:= pBrokerTZ;
        TimeStamp := pTimeStamp;
end;

destructor TDateSeekEntry.Destroy;
begin

end;

function TDateSeekEntry.getLastAccessed 	: tDateTime;
begin
        exit(lastAccessed);
end;

function TDateSeekEntry.getSeekDate		: TDateTime;
begin
        exit(seekDate);
end;

function TDateSeekEntry.getSeekTimeStr		: AnsiString;
begin
        exit(seekTimeStr);
end;

function TDateSeekEntry.getSeekTimeZone		: AnsiString;
begin
        exit(seekTimeZone);
end;

function TDateSeekEntry.getTimeStamp		: LongInt;
begin
	setAccessed;
        exit(Timestamp);
end;

end.

